
#include <map2.h>
#include <1905_tlvs.h>
#include <cmdu.h>
#include <i1905_wsc.h>

#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>
#include <libubox/uloop.h>
#include <libubox/ustream.h>
#include <libubox/utils.h>

#include <libubus.h>
#include <uci.h>
#include <easy/easy.h>

#include "utils/utils.h"
#include "utils/debug.h"
#include "config.h"
#include "cntlr.h"
#include "cntlr_cmdu.h"
#include "cntlr_tlv.h"
#include "cmdu_validate.h"

static int check_serialized_tlv(struct tlv *t, uint16_t len)
{
	uint8_t *tv_data;
	uint16_t tlv_len;

	if (!t)
		return -1;

	tlv_len = tlv_length(t);

	if (tlv_len != len)
		return -1;

	tv_data = (uint8_t *)t->data;
	if (!tv_data)
		return -1;

	return 0;
}

static int check_profile2_ap_cap_tlv(struct tlv *t)
{
	/* 4 bytes */
	return check_serialized_tlv(t,
		sizeof(struct tlv_profile2_ap_cap));
}

/* Check AP Radio Advanced Capabilities TLV */
static int check_ap_radio_adv_cap_tlv(struct tlv *t)
{
	/* radio (6) + cap (1 byte) */
	return check_serialized_tlv(t,
		sizeof(struct tlv_ap_radio_adv_cap));
}

/* Check MultiAP Profile TLV */
static int check_map_profile_tlv(struct tlv *t)
{
	/* profile (1 byte) */
	return check_serialized_tlv(t,
				sizeof(struct tlv_map_profile));
}

static int check_service_tlv(struct tlv *t)
{
	int offset = 0;
	uint8_t num_services;
	uint8_t *tv_data;
	uint16_t tlv_len;

	if (!t)
		return -1;

	tlv_len = tlv_length(t);
	/* at least 1 byte: num_services */
	if (tlv_len < 1)
		return -1;

	tv_data = (uint8_t *)t->data;
	if (!tv_data)
		return -1;

	num_services = tv_data[offset++];

	/* services (num_services bytes) */
	if (offset + num_services > tlv_len)
		return -1;

	return 0;
}

/* Check SupportedService TLV */
static int check_supported_service_tlv(struct tlv *t)
{
	return check_service_tlv(t);
}

/* Check SearchedService TLV */
static int check_searched_service_tlv(struct tlv *t)
{
	return check_service_tlv(t);
}

bool validate_channel_scan_report(struct cmdu_buff *cmdu, struct tlv *tv_tsp[][16],
		struct tlv *tv_scan[], int *num)
{
	int i, j;
	int ret = 0;
	struct tlv_policy d_policy_tsp[] = {
		[0] = {
			.type = MAP_TLV_TIMESTAMP,
			.present = TLV_PRESENT_ONE,
			.minlen = 1
		}
	};
	struct tlv_policy d_policy_scan[] = {
		[0] = {
			.type = MAP_TLV_CHANNEL_SCAN_RES,
			.present = TLV_PRESENT_MORE,
			.minlen = 9
		}
	};

	ret = cmdu_parse_tlvs(cmdu, tv_tsp, d_policy_tsp, 1);
	ret |= cmdu_parse_tlv_single(cmdu, tv_scan, d_policy_scan, num);
	if (ret) {
		dbg("%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv_tsp[0][0]) {
		dbg("%s: No TIMESTAMP_TLV received!\n", __func__);
		return false;
	}

	if (!tv_scan[0]) {
		dbg("%s: No RESULT_TLV received!\n", __func__);
		return false;
	}

	if (tv_tsp[0][0]) {
		int offset = 0;
		int tsp_len = 0;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv_tsp[0][0]);

		t = (uint8_t *)tv_tsp[0][0]->data;
		if (!t)
			return false;

		/* offset within provided minlen(1) */
		tsp_len = t[offset];
		offset += 1;
		if ((offset + tsp_len) > tlv_len)
			return false;
	}

	for (i = 0; i < *num; i++) {
		int offset = 0;
		uint8_t status = 0x00;
		uint8_t *t = NULL;
		uint16_t tlv_len = tlv_length(tv_scan[i]);
		int ts_len = 0;
		uint16_t num_nb = 0;

		t = (uint8_t *)tv_scan[i]->data;
		if (!t)
			return false;

		/* offset within provided minlen(9) */
		offset += 6;	/* radio id */
		offset += 1;	/* opclass */
		offset += 1;	/* channel */
		status = t[offset];
		offset += 1;	/* status */

		if (status == CH_SCAN_STATUS_SUCCESS) {

			if ((offset + 1) > tlv_len)
				return false;

			ts_len = t[offset];
			offset += 1;	/* timestamp len */
			if ((offset + ts_len + 4) > tlv_len)
				return false;

			offset += ts_len;	/* timestamp */
			offset += 1;		/* utilization */
			offset += 1;		/* noise */
			num_nb = BUF_GET_BE16(t[offset]);
			offset += 2;		/* num neightbors */

			for (j = 0; j < num_nb; j++) {
				int ssid_len = 0;
				int bw_len = 0;
				uint8_t info = 0x00;

				if ((offset + 7) > tlv_len)
					return false;

				offset += 6;	/* bssid */
				ssid_len = t[offset];
				offset += 1;	/* ssid len */
				if ((offset + ssid_len + 2) > tlv_len)
					return false;

				offset += ssid_len;	/* ssid */
				offset += 1;		/* rcpi */
				bw_len = t[offset];
				offset += 1;		/* bw len */
				if ((offset + bw_len + 1) > tlv_len)
					return false;

				offset += bw_len;	/* bandwidth */
				info = t[offset];
				offset += 1;		/* info */

				if (info & CH_SCAN_RESULT_BSSLOAD_PRESENT) {
					if ((offset + 3) > tlv_len)
						return false;

					offset += 1;	/* channel utilization */
					offset += 2;	/* station count */
				}
			}

			if ((offset + 5) > tlv_len)
				return false;

			offset += 4;	/* total scan duration */
			offset += 1;	/* scan type */
		}
	}

	return true;
}

bool validate_topology_response(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	trace("%s: --->\n", __func__);

	int ret = 0;
#ifdef EASYMESH_VENDOR_EXT
	int num = 0;
#endif
	struct tlv_policy a_policy[] = {
		[0] = { .type = TLV_TYPE_DEVICE_INFORMATION_TYPE,
				.present = TLV_PRESENT_ONE,
		},
		[1] = { .type = TLV_TYPE_DEVICE_BRIDGING_CAPABILITIES,
				.present = TLV_PRESENT_OPTIONAL_MORE,
		},
		[2] = { .type = TLV_TYPE_NON_1905_NEIGHBOR_DEVICE_LIST,
				.present = TLV_PRESENT_OPTIONAL_MORE,
		},
		[3] = { .type = TLV_TYPE_NEIGHBOR_DEVICE_LIST,
				.present = TLV_PRESENT_OPTIONAL_MORE,
		},
		[4] = { .type = TLV_TYPE_POWER_OFF_INTERFACE,
				.present = TLV_PRESENT_OPTIONAL_MORE,
		},
		[5] = { .type = TLV_TYPE_L2_NEIGHBOR_DEVICE,
				.present = TLV_PRESENT_OPTIONAL_MORE,
		},
		[6] = { .type = MAP_TLV_SUPPORTED_SERVICE,
				.present = TLV_PRESENT_OPTIONAL_ONE,
				.minlen = 1, /* num_services(==0) */
		},
		[7] = { .type = MAP_TLV_AP_OPERATIONAL_BSS,
				.present = TLV_PRESENT_ONE,
				.minlen = 1, /* tlv_ap_oper_bss:num_radio */
		},
		[8] = { .type = MAP_TLV_ASSOCIATED_CLIENTS,
				.present = TLV_PRESENT_OPTIONAL_ONE,
				.minlen = 1, /* tlv_assoc_client:num_bss */
		},
		[9] = { .type = MAP_TLV_MULTIAP_PROFILE,
				.present = TLV_PRESENT_ONE,
				.minlen = 1, /* tlv_map_profile:profile */
				.maxlen = 1,
		},
#ifdef EASYMESH_VENDOR_EXT
		[10] = { .type = TLV_TYPE_VENDOR_SPECIFIC,
				 .present = TLV_PRESENT_OPTIONAL_MORE,
				 .minlen = 3, /* tlv_vendor_bbss:oui+num_radios */
		}
#endif
	};

	trace("parsing topology response |" MACFMT "|CMDU: topology response\n",
			MAC2STR(cmdu->origin));

	ret = cmdu_parse_tlvs(cmdu, tv, a_policy, 11);
	if (ret) {
		dbg("%s: parse_tlv failed\n", __func__);
		return false;
	}

	/* Parse SupportedService TLV */
	if (check_supported_service_tlv(tv[6][0]))
		return false;

	/* MAP_TLV_AP_OPERATIONAL_BSS */
	if (tv[7][0]) {
		struct tlv_ap_oper_bss *tlv;
		uint8_t *tv_data;
		uint16_t tlv_len = 0;
		int i, offset = 0;

		tlv_len = tlv_length(tv[7][0]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_ap_oper_bss *)tv[7][0]->data;
		if (!tlv)
			return false;

		tv_data = (uint8_t *)tlv;

		/* num_radio (1 byte) */
		if (offset + 1 > tlv_len)
			return false;

		offset += 1;

		for (i = 0; i < tlv->num_radio; i++) {
			uint8_t num_bss = 0;
			int j;

			/* macaddr (6 bytes) */
			if (offset + 6 > tlv_len)
				return false;

			offset += 6;

			/* num_bss (1 byte) */
			if (offset + 1 > tlv_len)
				return false;

			memcpy(&num_bss, &tv_data[offset], 1);

			offset += 1;

			for (j = 0; j < num_bss; j++) {
				uint8_t ssidlen = 0;

				/* macaddr (6 bytes) */
				if (offset + 6 > tlv_len)
					return false;

				offset += 6;

				/* ssidlen (1 byte) */
				if (offset + 1 > tlv_len)
					return false;

				memcpy(&ssidlen, &tv_data[offset], 1);

				offset += 1;

				/* ssid (ssidlen bytes) */
				if (offset + ssidlen > tlv_len)
					return false;

				offset += ssidlen;
			}
		}
	}

	/* MAP_TLV_ASSOCIATED_CLIENTS */
	if (tv[8][0]) {
		uint16_t tlv_len = 0;
		struct tlv_assoc_client *tlv;
		uint8_t *tv_data;
		int i, offset = 0;

		tlv_len = tlv_length(tv[8][0]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_assoc_client *)tv[8][0]->data;
		if (!tlv)
			return false;

		tv_data = (uint8_t *)tlv;

		/* num_bss (1 byte) */
		if (offset + 1 > tlv_len)
			return false;

		offset += 1;

		for (i = 0; i < tlv->num_bss; i++) {
			uint16_t num_client = 0;

			/* radio id (6 bytes) */
			if (offset + 6 > tlv_len)
				return false;

			offset += 6;

			/* num_client (2 bytes) */
			if (offset + 2 > tlv_len)
				return false;

			num_client = BUF_GET_BE16(tv_data[offset]);

			offset += 2;

			 /* num_client * (macaddr + conntime) */
			if (offset + (num_client * 8) > tlv_len)
				return false;

			offset += (num_client * 8);
		}
	}

	/* Parse MultiAP Profile TLV */
	if (check_map_profile_tlv(tv[9][0]))
		return false;

#ifdef EASYMESH_VENDOR_EXT
	/* TLV_TYPE_VENDOR_SPECIFIC */
	while (tv[10][num]) {
		uint16_t tlv_len = 0;
		struct tlv_vendor_bbss *tlv;
		int i, offset = 0;
		uint8_t oui2[3];  /* TODO: use the same vendor oui-type */

		tlv_len = tlv_length(tv[10][num]);
		if (!tlv_len)
			return false;

		tlv = (struct tlv_vendor_bbss *)tv[10][num]->data;
		if (!tlv)
			return false;

		/* oui (3 bytes) */
		if (offset + 3 > tlv_len)
			return false;

		memcpy(oui2, EASYMESH_VENDOR_EXT_OUI, 3);
		oui2[2]++;
		if (memcmp(tlv->oui, oui2, 3)) {
			num++;
			continue;
		}

		offset += 3;

		/* num_radios (1 byte) */
		if (offset + 1 > tlv_len)
			return false;

		offset += 1;

		for (i = 0; i < tlv->num_radios; i++) {
			uint8_t num_bss = 0;

			/* macaddr (6 bytes) */
			if (offset + 6 > tlv_len)
				return false;

			offset += 6;

			/* num_bss (1 byte) */
			if (offset + 1 > tlv_len)
				return false;

			memcpy(&num_bss, &tv[10][num]->data[offset], 1);

			offset += 1;

			/* bss macaddrs (num_bss * 6 bytes) */
			if (offset + (num_bss * 6) > tlv_len)
				return false;

			offset += num_bss * 6;
		}

		num++;
	}
#endif /*EASYMESH_VENDOR_EXT*/
	return true;
}

#define ATTR_MSG_TYPE   (0x1022)
#define MSG_TYPE_M1	 (0x04)
static int validate_wsc_m1(uint8_t *m1, uint16_t m1_size)
{
	uint8_t *data;
	uint8_t *m1_end;
	int ret = -1;

	if (!m1 || !m1_size)
		return -1;

	data = m1;
	m1_end = m1 + m1_size;

	while (abs(data - m1) < m1_size - 4) {
		uint16_t attr_type;
		uint16_t attr_len;

		attr_type = buf_get_be16(data);
		data += 2;
		attr_len = buf_get_be16(data);
		data += 2;

		if (data + attr_len > m1_end) {
			dbg("%s: parse_wsc_m1 failed\n", __func__);
			ret = -1;
			break;
		}

		if (attr_type == ATTR_MSG_TYPE) {
			if (attr_len != 1) {
				ret = -1;
				break;
			}
			if (*data == MSG_TYPE_M1)
				ret = 0;
		}

		data += attr_len;
	}

	/* 0 if msg type is M1 & data never goes OOB */
	return ret;
}

/* Check WSC TLV (containing M1) */
static int check_wsc_tlv(struct tlv *t)
{
	uint8_t *tv_data;
	uint16_t tlv_len;

	if (!t)
		return -1;

	tlv_len = tlv_length(t);
	if (!tlv_len)
		return -1;

	tv_data = (uint8_t *)t->data;
	if (!tv_data)
		return -1;

	return validate_wsc_m1(tv_data, tlv_len);
}

/* Check AP Radio Basic Capabilities TLV */
static int check_ap_radio_basic_cap_tlv(struct tlv *t)
{
	uint8_t *tv_data;
	uint16_t tlv_len = 0;
	struct tlv_ap_radio_basic_cap *tlv;
	int i, offset = 0;

	tlv_len = tlv_length(t);
	if (!tlv_len)
		return -1;

	tlv = (struct tlv_ap_radio_basic_cap *) t->data;
	if (!tlv)
		return -1;

	tv_data = (uint8_t *)tlv;

	/* radio (6 bytes), max_bssnum (1), num_opclass (1) */
	if (offset + sizeof(*tlv) > tlv_len)
		return -1;

	offset += sizeof(*tlv);

	for (i = 0; i < tlv->num_opclass; i++) {
		struct ap_radio_basic_cap_opclass *op =
			(struct ap_radio_basic_cap_opclass *)&tv_data[offset];

		/* classid (1), max_txpower(1), num_nonomp_channel (1) */
		if (offset + sizeof(*op) > tlv_len)
			return -1;

		offset += sizeof(*op);

		/* nonop_channel (num_nonop_channel bytes) */
		if (offset + op->num_nonop_channel > tlv_len)
			return -1;

		offset += op->num_nonop_channel;
	}

	return 0;
}

bool validate_ap_autoconfig_wsc(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	struct tlv_policy a_policy[] = {
		[0] = { .type = MAP_TLV_AP_RADIO_BASIC_CAPABILITIES,
				.present = TLV_PRESENT_ONE,
				.minlen = 8, /* tlv_ap_radio_basic_cap */
		},
		[1] = { .type = MAP_TLV_PROFILE2_AP_CAPABILITY,
				.present = TLV_PRESENT_ONE,
				.minlen = 4, /* tlv_profile2_ap_cap */
				.maxlen = 4
		},
		[2] = { .type = MAP_TLV_AP_RADIO_ADV_CAPABILITY,
				.present = TLV_PRESENT_ONE,
				.minlen = 7, /* tlv_ap_radio_adv_cap */
				.maxlen = 7
		},
		[3] = { .type = TLV_TYPE_WSC,
				.present = TLV_PRESENT_ONE
		}
	};
	int ret = 0;

	trace("%s |" MACFMT "|CMDU: ap autoconfig response\n",
		  __func__, MAC2STR(cmdu->origin));

	ret = cmdu_parse_tlvs(cmdu, tv, a_policy, 4);
	if (ret) {
		dbg("%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0] || !tv[1][0] || !tv[2][0] || !tv[3][0]) {
		dbg("%s: Missing one or more mandatory TLV!\n", __func__);
		return false;
	}

	/* Parse AP Radio Basic Capabilities TLV */
	if (check_ap_radio_basic_cap_tlv(tv[0][0]))
		return false;

	/* Parse Profile-2 AP Capability TLV */
	if (check_profile2_ap_cap_tlv(tv[1][0]))
		return false;

	/* Parse AP Radio Advanced Capabilities TLV */
	if (check_ap_radio_adv_cap_tlv(tv[2][0]))
		return false;

	/* Parse WSC TLV (containing M1) */
	if (check_wsc_tlv(tv[3][0]))
		return false;

	return true;
}

/* 0: 1905.1 AL MAC address type TLV
 * 1: SearchedRole TLV
 * 2: AutoconfigFreqBand TLV
 * 3: SupportedService TLV
 * 4: SearchedService TLV
 * 5: MultiAP Profile TLV
 **/
bool validate_ap_autoconfig_search(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	struct tlv_policy a_policy[] = {
		[0] = { .type = TLV_TYPE_AL_MAC_ADDRESS_TYPE,
				.present = TLV_PRESENT_ONE,
				.minlen = 6, /* macaddr */
				.maxlen = 6
		},
		[1] = { .type = TLV_TYPE_SEARCHED_ROLE,
				.present = TLV_PRESENT_ONE,
				.minlen = 1, /* tlv_searched_role */
				.maxlen = 1
		},
		[2] = { .type = TLV_TYPE_AUTOCONFIG_FREQ_BAND,
				.present = TLV_PRESENT_ONE,
				.minlen = 1, /* tlv_autoconfig_band */
				.maxlen = 1
		},
		[3] = { .type = MAP_TLV_SUPPORTED_SERVICE,
				.present = TLV_PRESENT_OPTIONAL_ONE,
				.minlen = 1, /* num of services */
		},
		[4] = { .type = MAP_TLV_SEARCHED_SERVICE,
				.present = TLV_PRESENT_OPTIONAL_ONE,
				.minlen = 1, /* num of services */
		},
		[5] = { .type = MAP_TLV_MULTIAP_PROFILE,
				.present = TLV_PRESENT_ONE,
				.minlen = 1, /* tlv_map_profile */
				.maxlen = 1
		}
	};
	int ret;

	trace("%s |" MACFMT "|CMDU: ap autoconfig search\n",
		  __func__, MAC2STR(cmdu->origin));

	ret = cmdu_parse_tlvs(cmdu, tv, a_policy, 6);
	if (ret) {
		dbg("%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (tv[3][0] && check_supported_service_tlv(tv[3][0]))
		return false;

	if (tv[4][0] && check_searched_service_tlv(tv[4][0]))
		return false;

	return true;
}

/* 0: SupportedRole TLV
 * 1: SupportedFreqBand TLV
 * 2: SupportedService TLV
 * 3: MultiAP Profile TLV
 **/
bool validate_ap_autoconfig_response(struct cmdu_buff *cmdu, struct tlv *tv[][16])
{
	struct tlv_policy a_policy[] = {
		[0] = { .type = TLV_TYPE_SUPPORTED_ROLE,
				.present = TLV_PRESENT_ONE,
				.minlen = 1, /* tlv_supported_role */
				.maxlen = 1,
		},
		[1] = { .type = MAP_TLV_SUPPORTED_SERVICE,
				.present = TLV_PRESENT_OPTIONAL_ONE,
				.minlen = 1, /* num of services */
		},
		[2] = { .type = MAP_TLV_MULTIAP_PROFILE,
				.present = TLV_PRESENT_ONE,
				.minlen = 1, /* tlv_map_profile */
				.maxlen = 1
		},
		[3] = { .type = TLV_TYPE_SUPPORTED_FREQ_BAND,
				.present = TLV_PRESENT_ONE,
				.minlen = 1, /* tlv_supported_band */
				.maxlen = 1
		}
	};
	int ret;

	trace("%s |" MACFMT "|CMDU: ap autoconfig response\n",
		  __func__, MAC2STR(cmdu->origin));

	ret = cmdu_parse_tlvs(cmdu, tv, a_policy, 4);
	if (ret) {
		dbg("%s: parse_tlv failed\n", __func__);
		return false;
	}

	if (!tv[0][0] || !tv[2][0] || !tv[3][0]) {
		dbg("%s: Missing one or more mandatory TLV!\n", __func__);
		return false;
	}

	if (tv[1][0] && check_supported_service_tlv(tv[1][0]))
		return false;

	return true;
}
